#ifndef _layerExText_hpp_
#define _layerExText_hpp_

#include <windows.h>
#include <gdiplus.h>
using namespace Gdiplus;

#include <vector>
using namespace std;

#include "layerExBase.hpp"

/**
 * GDIPlus ŗLp
 */
struct GdiPlus {
	/**
	 * vCx[gtHg̒ǉ
	 * @param fontFileName tHgt@C
	 */
	static void addPrivateFont(const tjs_char *fontFileName);

	/**
	 * tHgt@~[擾
	 * @param privateOnly true ȂvCx[gtHĝݎ擾
	 */
	static tTJSVariant getFontList(bool privateOnly);
};

/**
 * tHg
 */
class FontInfo {
	friend class LayerExDraw;
	friend class Path;

protected:
	FontFamily *fontFamily; //< tHgt@~[
	ttstr familyName;
	REAL emSize; //< tHgTCY 
	INT style; //< tHgX^C
        bool gdiPlusUnsupportedFont; //< GDI+T|[gtHg
        bool forceSelfPathDraw; // OpX`拭
        mutable bool propertyModified;
        mutable REAL ascent;
        mutable REAL descent;
        mutable REAL lineSpacing;
        mutable REAL ascentLeading;
        mutable REAL descentLeading;

	/**
	 * tHg̃NA
	 */
	void clear();

        OUTLINETEXTMETRIC *createFontMetric(void) const;


public:

	FontInfo();
	/**
	 * RXgN^
	 * @param familyName tHgt@~[
	 * @param emSize tHg̃TCY
	 * @param style tHgX^C
	 */
	FontInfo(const tjs_char *familyName, REAL emSize, INT style);
	FontInfo(const FontInfo &orig);

	/**
	 * fXgN^
	 */
	virtual ~FontInfo();

	void setFamilyName(const tjs_char *familyName);
	const tjs_char *getFamilyName() { return familyName.c_str(); }
	void setEmSize(REAL emSize) { this->emSize = emSize; propertyModified = true; }
	REAL getEmSize() {  return emSize; }
	void setStyle(INT style) { this->style = style; propertyModified = true; }
	INT getStyle() { return style; }
        void setForceSelfPathDraw(bool state);
        bool getForceSelfPathDraw(void) const;
        bool getSelfPathDraw(void) const;

        void updateSizeParams(void) const;
	REAL getAscent() const;
	REAL getDescent() const;
	REAL getAscentLeading() const;
	REAL getDescentLeading() const;
	REAL getLineSpacing() const;
};

/**
 * `OϏ
 */
class Appearance {
	friend class LayerExDraw;
public:
	// `
	struct DrawInfo{
		int type;   // 0:uV 1:y
		void *info; // IuWFNg
		REAL ox; //< \ItZbg
		REAL oy; //< \ItZbg
		DrawInfo() : ox(0), oy(0), type(0), info(NULL) {}
		DrawInfo(REAL ox, REAL oy, Pen *pen) : ox(ox), oy(oy), type(0), info(pen) {}
		DrawInfo(REAL ox, REAL oy, Brush *brush) : ox(ox), oy(oy), type(1), info(brush) {}
		DrawInfo(const DrawInfo &orig) {
			ox = orig.ox;
			oy = orig.oy;
			type = orig.type;
			if (orig.info) {
				switch (type) {
				case 0:
					info = (void*)((Pen*)orig.info)->Clone();
					break;
				case 1:
					info = (void*)((Brush*)orig.info)->Clone();
					break;
				}
			} else {
				info = NULL;
			}
		}
		virtual ~DrawInfo() {
			if (info) {
				switch (type) {
				case 0:
					delete (Pen*)info;
					break;
				case 1:
					delete (Brush*)info;
					break;
				}
			}
		}	
	};
	vector<DrawInfo>drawInfos;

public:
	Appearance();
	virtual ~Appearance();

	/**
	 * ̃NA
	 */
	void clear();
	
	/**
	 * uV̒ǉ
	 * @param colorOrBrush ARGBFw܂̓uVij
	 * @param ox \ItZbgX
	 * @param oy \ItZbgY
	 */
	void addBrush(tTJSVariant colorOrBrush, REAL ox=0, REAL oy=0);
	
	/**
	 * y̒ǉ
	 * @param colorOrBrush ARGBFw܂̓uVij
	 * @param widthOrOption y܂̓yij
	 * @param ox \ItZbgX
	 * @param oy \ItZbgY
	 */
	void addPen(tTJSVariant colorOrBrush, tTJSVariant widthOrOption, REAL ox=0, REAL oy=0);

protected:
	/**
	 * LineCap̎擾
	 */
	bool getLineCap(tTJSVariant &in, LineCap &cap, CustomLineCap* &custom, REAL pw);
	vector<CustomLineCap*>customLineCaps;
};


/**
 * `OϏ
 */
class Path {
	friend class LayerExDraw;
public:
	Path();
	virtual ~Path();
	void startFigure();
	void closeFigure();
	void drawArc(REAL x, REAL y, REAL width, REAL height, REAL startAngle, REAL sweepAngle);
	void drawBezier(REAL x1, REAL y1, REAL x2, REAL y2, REAL x3, REAL y3, REAL x4, REAL y4);
	void drawBeziers(tTJSVariant points);
	void drawClosedCurve(tTJSVariant points);
	void drawClosedCurve2(tTJSVariant points, REAL tension);
	void drawCurve(tTJSVariant points);
	void drawCurve2(tTJSVariant points, REAL tension);
	void drawCurve3(tTJSVariant points, int offset, int numberOfSegments, REAL tension);
	void drawPie(REAL x, REAL y, REAL width, REAL height, REAL startAngle, REAL sweepAngle);
	void drawEllipse(REAL x, REAL y, REAL width, REAL height);
	void drawLine(REAL x1, REAL y1, REAL x2, REAL y2);
	void drawLines(tTJSVariant points);
	void drawPolygon(tTJSVariant points);
	void drawRectangle(REAL x, REAL y, REAL width, REAL height);
	void drawRectangles(tTJSVariant rects);
protected:
	GraphicsPath path;
};

/*
 * AEgCx[X̃eLXg`惁\bh̒ǉ
 */
class LayerExDraw : public layerExBase
{
protected:
	// ێp
	GeometryT width, height;
	BufferT   buffer;
	PitchT    pitch;
	GeometryT clipLeft, clipTop, clipWidth, clipHeight;
	
	/// CQƂrbg}bv
	Bitmap *bitmap;
	/// Cɑ΂ĕ`悷ReLXg
	Graphics *graphics;

	// Transform w
	Matrix transform;
	Matrix viewTransform;
	Matrix calcTransform;

protected:
	// `X[WOw
	SmoothingMode smoothingMode;
	// drawString ̃A`GCAXw
	TextRenderingHint textRenderingHint;

public:
	int getSmoothingMode() {
		return (int)smoothingMode;
	}
	void setSmoothingMode(int mode) {
		smoothingMode = (SmoothingMode)mode;
	}

	int getTextRenderingHint() {
		return (int)textRenderingHint;
	}
	void setTextRenderingHint(int hint) {
		textRenderingHint = (TextRenderingHint)hint;
	}

protected:
	/// `eL^p^t@C
	HDC metaHDC;
	HGLOBAL metaBuffer;
	IStream *metaStream;
	Metafile *metafile;
	Graphics *metaGraphics;

	bool updateWhenDraw;
	void updateRect(RectF &rect);
	
public:
	void setUpdateWhenDraw(int updateWhenDraw) {
		this->updateWhenDraw = updateWhenDraw != 0;
	}
	int getUpdateWhenDraw() { return updateWhenDraw ? 1 : 0; }

	inline operator Image*() const { return (Image*)bitmap; }
	inline operator Bitmap*() const { return bitmap; }
	inline operator Graphics*() const { return graphics; }
	inline operator const Image*() const { return (const Image*)bitmap; }
	inline operator const Bitmap*() const { return bitmap; }
	inline operator const Graphics*() const { return graphics; }
	
	template <class T>
	struct BridgeFunctor {
		T* operator()(LayerExDraw *p) const {
			return (T*)*p;
		}
	};

public:	
	LayerExDraw(DispatchT obj);
	~LayerExDraw();
	virtual void reset();

	// ------------------------------------------------------------------
	// `p[^w
	// ------------------------------------------------------------------

protected:
	void updateViewTransform();
	void updateTransform();
	
public:
	/**
	 * \gXtH[̎w
	 */
	void setViewTransform(const Matrix *transform);
	void resetViewTransform();
	void rotateViewTransform(REAL angle);
	void scaleViewTransform(REAL sx, REAL sy);
	void translateViewTransform(REAL dx, REAL dy);
	
	/**
	 * gXtH[̎w
	 * @param matrix gXtH[}gbNX
	 */
	void setTransform(const Matrix *transform);
	void resetTransform();
	void rotateTransform(REAL angle);
	void scaleTransform(REAL sx, REAL sy);
	void translateTransform(REAL dx, REAL dy);

	// ------------------------------------------------------------------
	// `惁\bhQ
	// ------------------------------------------------------------------

protected:

	/**
	 * pX̍XV̈擾
	 * @param app \\
	 * @param path `悷pX
	 * @return XV̈
	 */
	RectF getPathExtents(const Appearance *app, const GraphicsPath *path);

	/**
	 * pX̕`p
	 * @param graphics `
	 * @param pen `py
	 * @param matrix `ʒupmatrix
	 * @param path `e
	 */
	void draw(Graphics *graphics, const Pen *pen, const Matrix *matrix, const GraphicsPath *path);

	/**
	 * h̕`p
	 * @param graphics `
	 * @param brush `puV
	 * @param matrix `ʒupmatrix
	 * @param path `e
	 */
	void fill(Graphics *graphics, const Brush *brush, const Matrix *matrix, const GraphicsPath *path);
	
	/**
	 * pX̕`
	 * @param app AsAX
	 * @param path `悷pX
	 * @return XV̈
	 */
	RectF _drawPath(const Appearance *app, const GraphicsPath *path);

        /**
         * OtAEgC̎擾
         * @param font tHg
         * @param offset ItZbg
         * @param path OtopX
         * @param glyph `悷Ot
         */
        void getGlyphOutline(const FontInfo *font, PointF &offset, GraphicsPath *path, UINT glyph);

        /*
         * eLXgAEgC̎擾
         * @param font tHg
         * @param offset ItZbg
         * @param path OtopX
         * @param text `悷eLXg
         */
         void getTextOutline(const FontInfo *font, PointF &offset, GraphicsPath *path, ttstr text);

public:
	/**
	 * ʂ̏
	 * @param argb F
	 */
	void clear(ARGB argb);

	/**
	 * pX̕`
	 * @param app AsAX
	 * @param path pX
	 */
	RectF drawPath(const Appearance *app, const Path *path);
	
	/**
	 * ~ʂ̕`
	 * @param app AsAX
	 * @param x W
	 * @param y W
	 * @param width 
	 * @param height c
	 * @param startAngle v~ʊJnʒu
	 * @param sweepAngle `px
	 * @return XV̈
	 */
	RectF drawArc(const Appearance *app, REAL x, REAL y, REAL width, REAL height, REAL startAngle, REAL sweepAngle);

	/**
	 * ~̕`
	 * @param app AsAX
	 * @param x W
	 * @param y W
	 * @param width 
	 * @param height c
	 * @param startAngle v~ʊJnʒu
	 * @param sweepAngle `px
	 * @return XV̈
	 */
	RectF drawPie(const Appearance *app, REAL x, REAL y, REAL width, REAL height, REAL startAngle, REAL sweepAngle);
	
	/**
	 * xWFȐ̕`
	 * @param app AsAX
	 * @param x1
	 * @param y1
	 * @param x2
	 * @param y2
	 * @param x3
	 * @param y3
	 * @param x4
	 * @param y4
	 * @return XV̈
	 */
	RectF drawBezier(const Appearance *app, REAL x1, REAL y1, REAL x2, REAL y2, REAL x3, REAL y3, REAL x4, REAL y4);

	/**
	 * AxWFȐ̕`
	 * @param app AsAX
	 * @param points _̔z
	 * @return XV̈
	 */
	RectF drawBeziers(const Appearance *app, tTJSVariant points);

	/**
	 * Closed cardinal spline ̕`
	 * @param app AsAX
	 * @param points _̔z
	 * @return XV̈
	 */
	RectF drawClosedCurve(const Appearance *app, tTJSVariant points);

	/**
	 * Closed cardinal spline ̕`
	 * @param app AsAX
	 * @param points _̔z
	 * @pram tension tension
	 * @return XV̈
	 */
	RectF drawClosedCurve2(const Appearance *app, tTJSVariant points, REAL tension);

	/**
	 * cardinal spline ̕`
	 * @param app AsAX
	 * @param points _̔z
	 * @return XV̈
	 */
	RectF drawCurve(const Appearance *app, tTJSVariant points);

	/**
	 * cardinal spline ̕`
	 * @param app AsAX
	 * @param points _̔z
	 * @parma tension tension
	 * @return XV̈
	 */
	RectF drawCurve2(const Appearance *app, tTJSVariant points, REAL tension);

	/**
	 * cardinal spline ̕`
	 * @param app AsAX
	 * @param points _̔z
	 * @param offset
	 * @param numberOfSegment
	 * @param tension tension
	 * @return XV̈
	 */
	RectF drawCurve3(const Appearance *app, tTJSVariant points, int offset, int numberOfSegments, REAL tension);
	
	/**
	 * ȉ~̕`
	 * @param app AsAX
	 * @param x
	 * @param y
	 * @param width
	 * @param height
	 * @return XV̈
	 */
	RectF drawEllipse(const Appearance *app, REAL x, REAL y, REAL width, REAL height);

	/**
	 * ̕`
	 * @param app AsAX
	 * @param x1 n_XW
	 * @param y1 n_YW
	 * @param x2 I_XW
	 * @param y2 I_YW
	 * @return XV̈
	 */
	RectF drawLine(const Appearance *app, REAL x1, REAL y1, REAL x2, REAL y2);

	/**
	 * A̕`
	 * @param app AsAX
	 * @param points _̔z
	 * @return XV̈
	 */
	RectF drawLines(const Appearance *app, tTJSVariant points);

	/**
	 * p`̕`
	 * @param app AsAX
	 * @param points _̔z
	 * @return XV̈
	 */
	RectF drawPolygon(const Appearance *app, tTJSVariant points);
	
	/**
	 * `̕`
	 * @param app AsAX
	 * @param x
	 * @param y
	 * @param width
	 * @param height
	 * @return XV̈
	 */
	RectF drawRectangle(const Appearance *app, REAL x, REAL y, REAL width, REAL height);

	/**
	 * `̕`
	 * @param app AsAX
	 * @param rects `̔z
	 * @return XV̈
	 */
	RectF drawRectangles(const Appearance *app, tTJSVariant rects);

	/**
	 * ̕`
	 * @param font tHg
	 * @param app AsAX
	 * @param x `ʒuX
	 * @param y `ʒuY
	 * @param text `eLXg
	 * @return XV̈
	 */
	RectF drawPathString(const FontInfo *font, const Appearance *app, REAL x, REAL y, const tjs_char *text);

	/**
	 * ̕`(OpenTypePostScripttHgΉ)
	 * @param font tHg
	 * @param app AsAX
	 * @param x `ʒuX
	 * @param y `ʒuY
	 * @param text `eLXg
	 * @return XV̈
	 */
	RectF drawPathString2(const FontInfo *font, const Appearance *app, REAL x, REAL y, const tjs_char *text);

	// -------------------------------------------------------------------------------
	
	/**
	 * ̕`
	 * @param font tHg
	 * @param app AsAX
	 * @param x `ʒuX
	 * @param y `ʒuY
	 * @param text `eLXg
	 * @return XV̈
	 */
	RectF drawString(const FontInfo *font, const Appearance *app, REAL x, REAL y, const tjs_char *text);

	/**
	 * ̕`XV̈̎擾
	 * @param font tHg
	 * @param text `eLXg
	 * @return XV̈̎ left, top, width, height
	 */
	RectF measureString(const FontInfo *font, const tjs_char *text);

	/**
	 * ɂ҂Ɛڂ`̎擾
	 * @param font tHg
	 * @param text `eLXg
	 * @return ̈̎ left, top, width, height
	 */
	RectF measureStringInternal(const FontInfo *font, const tjs_char *text);

	/**
	 * ̕`XV̈̎擾(OpenTypePostScripttHgΉ)
	 * @param font tHg
	 * @param text `eLXg
	 * @return XV̈̎ left, top, width, height
	 */
	RectF measureString2(const FontInfo *font, const tjs_char *text);

	/**
	 * ɂ҂Ɛڂ`̎擾(OpenTypePostScripttHgΉ)
	 * @param font tHg
	 * @param text `eLXg
	 * @return ̈̎ left, top, width, height
	 */
	RectF measureStringInternal2(const FontInfo *font, const tjs_char *text);

	// -----------------------------------------------------------------------------
	
	/**
	 * 摜̕`BRs[͌摜 Bounds zʒuATCY Pixel wɂȂ܂B
	 * @param x Rs[挴_X
	 * @param y Rs[挴_Y
	 * @param image Rs[摜
	 * @return XV̈
	 */
	RectF drawImage(REAL x, REAL y, Image *src);

	/**
	 * 摜̋`Rs[
	 * @param dleft Rs[捶[
	 * @param dtop  Rs[[
	 * @param src Rs[摜
	 * @param sleft `̍[
	 * @param stop  `̏[
	 * @param swidth `̉
	 * @param sheight  `̏c
	 * @return XV̈
	 */
	RectF drawImageRect(REAL dleft, REAL dtop, Image *src, REAL sleft, REAL stop, REAL swidth, REAL sheight);

	/**
	 * 摜̊gkRs[
	 * @param dleft Rs[捶[
	 * @param dtop  Rs[[
	 * @param dwidth Rs[̉
	 * @param dheight  Rs[̏c
	 * @param src Rs[摜
	 * @param sleft `̍[
	 * @param stop  `̏[
	 * @param swidth `̉
	 * @param sheight  `̏c
	 * @return XV̈
	 */
	RectF drawImageStretch(REAL dleft, REAL dtop, REAL dwidth, REAL dheight, Image *src, REAL sleft, REAL stop, REAL swidth, REAL sheight);

	/**
	 * 摜̃AtBϊRs[
	 * @param src Rs[摜
	 * @param sleft `̍[
	 * @param stop  `̏[
	 * @param swidth `̉
	 * @param sheight  `̏c
	 * @param affine AtBp[^̎(true:ϊs, false:Ww), 
	 * @return XV̈
	 */
	RectF drawImageAffine(Image *src, REAL sleft, REAL stop, REAL swidth, REAL sheight, bool affine, REAL A, REAL B, REAL C, REAL D, REAL E, REAL F);

	// ------------------------------------------------
	// ^t@C
	// ------------------------------------------------

protected:

	/**
	 * L^̐
	 */
	void createRecord();

	/**
	 * L^̐
	 */
	void recreateRecord();
	
	/**
	 * L^̔j
	 */
	void destroyRecord();

	/**
	 * ĕ`p
	 */
	bool redraw(Image *image);
	
public:
	/**
	 * @param record `eL^邩ǂ
	 */
	void setRecord(bool record);

	/**
	 * @return record `eL^邩ǂ
	 */
	bool getRecord() {
		return metafile != NULL;
	}

	/**
	 * L^e Image ƂĎ擾
	 * @return  true
	 */
	Image *getRecordImage();
	
	/**
	 * L^e̍ĕ`
	 * @return ĕ`悵 true
	 */
	bool redrawRecord();

	/**
	 * L^e̕ۑ
	 * @param filename ۑt@C
	 * @return  true
	 */
	bool saveRecord(const tjs_char *filename);

	/**
	 * L^e̓ǂݍ
	 * @param filename ǂݍ݃t@C
	 * @return  true
	 */
	bool loadRecord(const tjs_char *filename);

	/**
	 * 摜̕ۑ
	 */
	static tjs_error TJS_INTF_METHOD saveImage(tTJSVariant *result,
											   tjs_int numparams,
											   tTJSVariant **param,
											   iTJSDispatch2 *objthis);
};

#endif
